import { dummyLogger } from 'ts-log';
import { exec } from 'child_process';
import path, { resolve, extname, join, delimiter, sep, dirname, isAbsolute, relative } from 'path';
import { quote } from 'shell-quote';
import { createNoopProfiler, createProfiler, getCachedDocumentNodeFromSchema, normalizeInstanceOrArray, normalizeOutputParam, normalizeConfig } from '@graphql-codegen/plugin-helpers';
import { codegen } from '@graphql-codegen/core';
import { AggregateError, isValidPath } from '@graphql-tools/utils';
import { print, GraphQLError } from 'graphql';
import { cosmiconfig, defaultLoaders } from 'cosmiconfig';
import { TypeScriptLoader } from 'cosmiconfig-typescript-loader';
import { env } from 'string-env-interpolation';
import yargs from 'yargs';
import { loadConfig } from 'graphql-config';
import { ApolloEngineLoader } from '@graphql-tools/apollo-engine-loader';
import { CodeFileLoader } from '@graphql-tools/code-file-loader';
import { GitLoader } from '@graphql-tools/git-loader';
import { GithubLoader } from '@graphql-tools/github-loader';
import { PrismaLoader } from '@graphql-tools/prisma-loader';
import { loadSchema as loadSchema$1, loadDocuments as loadDocuments$1 } from '@graphql-tools/load';
import { GraphQLFileLoader } from '@graphql-tools/graphql-file-loader';
import { JsonFileLoader } from '@graphql-tools/json-file-loader';
import { UrlLoader } from '@graphql-tools/url-loader';
import yaml from 'yaml';
import { createRequire } from 'module';
import fs, { promises, unlink, writeFileSync, readFileSync } from 'fs';
import { createHash } from 'crypto';
import { cpus } from 'os';
import { Listr } from 'listr2';
import isGlob from 'is-glob';
import debounce from 'debounce';
import logSymbols from 'log-symbols';
import mkdirp from 'mkdirp';
import inquirer from 'inquirer';
import chalk from 'chalk';
import detectIndent from 'detect-indent';
import { fetch } from '@whatwg-node/fetch';
import template from '@babel/template';
import generate$1 from '@babel/generator';
import * as t from '@babel/types';

async function getPluginByName(name, pluginLoader) {
  const possibleNames = [
    `@graphql-codegen/${name}`,
    `@graphql-codegen/${name}-template`,
    `@graphql-codegen/${name}-plugin`,
    `graphql-codegen-${name}`,
    `graphql-codegen-${name}-template`,
    `graphql-codegen-${name}-plugin`,
    `codegen-${name}`,
    `codegen-${name}-template`,
    name
  ];
  const possibleModules = possibleNames.concat(resolve(process.cwd(), name));
  for (const moduleName of possibleModules) {
    try {
      return await pluginLoader(moduleName);
    } catch (err) {
      if (err.code !== "MODULE_NOT_FOUND" && err.code !== "ERR_MODULE_NOT_FOUND") {
        throw new Error(
          `
              Unable to load template plugin matching '${name}'.
              Reason:
                ${err.message}
            `
        );
      }
    }
  }
  const possibleNamesMsg = possibleNames.map(
    (name2) => `
        - ${name2}
    `.trimRight()
  ).join("");
  throw new Error(
    `
        Unable to find template plugin matching '${name}'
        Install one of the following packages:

        ${possibleNamesMsg}
      `
  );
}

async function getPresetByName(name, loader) {
  const possibleNames = [
    `@graphql-codegen/${name}`,
    `@graphql-codegen/${name}-preset`,
    name,
    resolve(process.cwd(), name)
  ];
  for (const moduleName of possibleNames) {
    try {
      const loaded = await loader(moduleName);
      if (loaded && loaded.preset) {
        return loaded.preset;
      }
      if (loaded && loaded.default) {
        return loaded.default;
      }
      return loaded;
    } catch (err) {
      if (err.code !== "MODULE_NOT_FOUND" && err.code !== "ERR_MODULE_NOT_FOUND") {
        throw new Error(
          `Unable to load preset matching ${name}

              Unable to load preset matching '${name}'.
              Reason:
                ${err.message}
            `
        );
      }
    }
  }
  const possibleNamesMsg = possibleNames.map(
    (name2) => `
        - ${name2}
    `.trimRight()
  ).join("");
  throw new Error(
    `Unable to find preset matching ${name}

        Unable to find preset matching '${name}'
        Install one of the following packages:

        ${possibleNamesMsg}
      `
  );
}

let logger;
function getLogger() {
  return logger || dummyLogger;
}
useWinstonLogger();
function useWinstonLogger() {
  if (logger && logger.levels) {
    return;
  }
  logger = console;
}

let queue = [];
function debugLog(message, ...meta) {
  queue.push({
    message,
    meta
  });
}
function printLogs() {
  queue.forEach((log) => {
    getLogger().info(log.message, ...log.meta);
  });
  resetLogs();
}
function resetLogs() {
  queue = [];
}

const CodegenExtension = (api) => {
  api.loaders.schema.register(
    new CodeFileLoader({
      pluckConfig: {
        skipIndent: true
      }
    })
  );
  api.loaders.schema.register(new GitLoader());
  api.loaders.schema.register(new GithubLoader());
  api.loaders.schema.register(new ApolloEngineLoader());
  api.loaders.schema.register(new PrismaLoader());
  api.loaders.documents.register(
    new CodeFileLoader({
      pluckConfig: {
        skipIndent: true
      }
    })
  );
  api.loaders.documents.register(new GitLoader());
  api.loaders.documents.register(new GithubLoader());
  return {
    name: "codegen"
  };
};
async function findAndLoadGraphQLConfig(filepath) {
  const config = await loadConfig({
    filepath,
    rootDir: process.cwd(),
    extensions: [CodegenExtension],
    throwOnEmpty: false,
    throwOnMissing: false
  });
  if (isGraphQLConfig(config)) {
    return config;
  }
}
function isGraphQLConfig(config) {
  if (!config) {
    return false;
  }
  try {
    return config.getDefault().hasExtension("codegen");
  } catch (e) {
  }
  try {
    for (const projectName in config.projects) {
      if (config.projects.hasOwnProperty(projectName)) {
        const project = config.projects[projectName];
        if (project.hasExtension("codegen")) {
          return true;
        }
      }
    }
  } catch (e) {
  }
  return false;
}

const defaultSchemaLoadOptions = {
  assumeValidSDL: true,
  sort: true,
  convertExtensions: true,
  includeSources: true
};
const defaultDocumentsLoadOptions = {
  sort: true,
  skipGraphQLImport: true
};
async function loadSchema(schemaPointers, config) {
  try {
    const loaders = [
      new CodeFileLoader(),
      new GitLoader(),
      new GithubLoader(),
      new GraphQLFileLoader(),
      new JsonFileLoader(),
      new UrlLoader(),
      new ApolloEngineLoader(),
      new PrismaLoader()
    ];
    const schema = await loadSchema$1(schemaPointers, {
      ...defaultSchemaLoadOptions,
      loaders,
      ...config,
      ...config.config
    });
    return schema;
  } catch (e) {
    throw new Error(
      `
        Failed to load schema from ${Object.keys(schemaPointers).join(",")}:

        ${e.message || e}
        ${e.stack || ""}

        GraphQL Code Generator supports:
          - ES Modules and CommonJS exports (export as default or named export "schema")
          - Introspection JSON File
          - URL of GraphQL endpoint
          - Multiple files with type definitions (glob expression)
          - String in config file

        Try to use one of above options and run codegen again.

      `
    );
  }
}
async function loadDocuments(documentPointers, config) {
  const loaders = [
    new CodeFileLoader({
      pluckConfig: {
        skipIndent: true
      }
    }),
    new GitLoader(),
    new GithubLoader(),
    new GraphQLFileLoader()
  ];
  const ignore = [];
  for (const generatePath of Object.keys(config.generates)) {
    if (extname(generatePath) === "") {
      continue;
    }
    ignore.push(join(process.cwd(), generatePath));
  }
  try {
    const loadedFromToolkit = await loadDocuments$1(documentPointers, {
      ...defaultDocumentsLoadOptions,
      ignore,
      loaders,
      ...config,
      ...config.config
    });
    return loadedFromToolkit;
  } catch (error) {
    if (config.ignoreNoDocuments)
      return [];
    throw error;
  }
}

const { lstat } = promises;
function generateSearchPlaces(moduleName) {
  const extensions = ["json", "yaml", "yml", "js", "ts", "config.js"];
  const regular = extensions.map((ext) => `${moduleName}.${ext}`);
  const dot = extensions.filter((ext) => ext !== "config.js").map((ext) => `.${moduleName}rc.${ext}`);
  return [...regular.concat(dot), "package.json"];
}
function customLoader(ext) {
  function loader(filepath, content) {
    if (typeof process !== "undefined" && "env" in process) {
      content = env(content);
    }
    if (ext === "json") {
      return defaultLoaders[".json"](filepath, content);
    }
    if (ext === "yaml") {
      try {
        const result = yaml.parse(content, { prettyErrors: true, merge: true });
        return result;
      } catch (error) {
        error.message = `YAML Error in ${filepath}:
${error.message}`;
        throw error;
      }
    }
    if (ext === "js") {
      return defaultLoaders[".js"](filepath, content);
    }
    if (ext === "ts") {
      const tsLoader = TypeScriptLoader({ transpileOnly: true });
      return tsLoader(filepath, content);
    }
  }
  return loader;
}
async function loadCodegenConfig({
  configFilePath,
  moduleName,
  searchPlaces: additionalSearchPlaces,
  packageProp,
  loaders: customLoaders
}) {
  configFilePath = configFilePath || process.cwd();
  moduleName = moduleName || "codegen";
  packageProp = packageProp || moduleName;
  const cosmi = cosmiconfig(moduleName, {
    searchPlaces: generateSearchPlaces(moduleName).concat(additionalSearchPlaces || []),
    packageProp,
    loaders: {
      ".json": customLoader("json"),
      ".yaml": customLoader("yaml"),
      ".yml": customLoader("yaml"),
      ".js": customLoader("js"),
      ".ts": customLoader("ts"),
      noExt: customLoader("yaml"),
      ...customLoaders
    }
  });
  const pathStats = await lstat(configFilePath);
  return pathStats.isDirectory() ? cosmi.search(configFilePath) : cosmi.load(configFilePath);
}
async function loadContext(configFilePath) {
  const graphqlConfig = await findAndLoadGraphQLConfig(configFilePath);
  if (graphqlConfig) {
    return new CodegenContext({
      graphqlConfig
    });
  }
  const result = await loadCodegenConfig({ configFilePath });
  if (!result) {
    if (configFilePath) {
      throw new Error(
        `
        Config ${configFilePath} does not exist.

          $ graphql-codegen --config ${configFilePath}

        Please make sure the --config points to a correct file.
      `
      );
    }
    throw new Error(
      `Unable to find Codegen config file! 

        Please make sure that you have a configuration file under the current directory!
      `
    );
  }
  if (result.isEmpty) {
    throw new Error(
      `Found Codegen config file but it was empty! 

        Please make sure that you have a valid configuration file under the current directory!
      `
    );
  }
  return new CodegenContext({
    filepath: result.filepath,
    config: result.config
  });
}
function getCustomConfigPath(cliFlags) {
  const configFile = cliFlags.config;
  return configFile ? resolve(process.cwd(), configFile) : null;
}
function buildOptions() {
  return {
    c: {
      alias: "config",
      type: "string",
      describe: 'Path to GraphQL codegen YAML config file, defaults to "codegen.yml" on the current directory'
    },
    w: {
      alias: "watch",
      describe: "Watch for changes and execute generation automatically. You can also specify a glob expression for custom watch list.",
      coerce: (watch) => {
        if (watch === "false") {
          return false;
        }
        if (typeof watch === "string" || Array.isArray(watch)) {
          return watch;
        }
        return !!watch;
      }
    },
    r: {
      alias: "require",
      describe: "Loads specific require.extensions before running the codegen and reading the configuration",
      type: "array",
      default: []
    },
    o: {
      alias: "overwrite",
      describe: "Overwrites existing files",
      type: "boolean"
    },
    s: {
      alias: "silent",
      describe: "Suppresses printing errors",
      type: "boolean"
    },
    e: {
      alias: "errors-only",
      describe: "Only print errors",
      type: "boolean"
    },
    profile: {
      describe: "Use profiler to measure performance",
      type: "boolean"
    },
    p: {
      alias: "project",
      describe: "Name of a project in GraphQL Config",
      type: "string"
    },
    v: {
      alias: "verbose",
      describe: "output more detailed information about performed tasks",
      type: "boolean",
      default: false
    },
    d: {
      alias: "debug",
      describe: "Print debug logs to stdout",
      type: "boolean",
      default: false
    }
  };
}
function parseArgv(argv = process.argv) {
  return yargs(argv).options(buildOptions()).parse(argv);
}
async function createContext(cliFlags = parseArgv(process.argv)) {
  if (cliFlags.require && cliFlags.require.length > 0) {
    const relativeRequire = createRequire(process.cwd());
    await Promise.all(
      cliFlags.require.map(
        (mod) => import(relativeRequire.resolve(mod, {
          paths: [process.cwd()]
        }))
      )
    );
  }
  const customConfigPath = getCustomConfigPath(cliFlags);
  const context = await loadContext(customConfigPath);
  updateContextWithCliFlags(context, cliFlags);
  return context;
}
function updateContextWithCliFlags(context, cliFlags) {
  const config = {
    configFilePath: context.filepath
  };
  if (cliFlags.watch !== void 0) {
    config.watch = cliFlags.watch;
  }
  if (cliFlags.overwrite === true) {
    config.overwrite = cliFlags.overwrite;
  }
  if (cliFlags.silent === true) {
    config.silent = cliFlags.silent;
  }
  if (cliFlags.verbose === true || process.env.VERBOSE) {
    config.verbose = true;
  }
  if (cliFlags.debug === true || process.env.DEBUG) {
    config.debug = true;
  }
  if (cliFlags.errorsOnly === true) {
    config.errorsOnly = cliFlags.errorsOnly;
  }
  if (cliFlags["ignore-no-documents"] !== void 0) {
    config.ignoreNoDocuments = cliFlags["ignore-no-documents"] === true;
  }
  if (cliFlags["emit-legacy-common-js-imports"] !== void 0) {
    config.emitLegacyCommonJSImports = cliFlags["emit-legacy-common-js-imports"] === true;
  }
  if (cliFlags.project) {
    context.useProject(cliFlags.project);
  }
  if (cliFlags.profile === true) {
    context.useProfiler();
  }
  if (cliFlags.check === true) {
    context.enableCheckMode();
  }
  context.updateConfig(config);
}
class CodegenContext {
  constructor({
    config,
    graphqlConfig,
    filepath
  }) {
    this._checkMode = false;
    this._pluginContext = {};
    this.checkModeStaleFiles = [];
    this._config = config;
    this._graphqlConfig = graphqlConfig;
    this.filepath = this._graphqlConfig ? this._graphqlConfig.filepath : filepath;
    this.cwd = this._graphqlConfig ? this._graphqlConfig.dirpath : process.cwd();
    this.profiler = createNoopProfiler();
  }
  useProject(name) {
    this._project = name;
  }
  getConfig(extraConfig) {
    if (!this.config) {
      if (this._graphqlConfig) {
        const project = this._graphqlConfig.getProject(this._project);
        this.config = {
          ...project.extension("codegen"),
          schema: project.schema,
          documents: project.documents,
          pluginContext: this._pluginContext
        };
      } else {
        this.config = { ...this._config, pluginContext: this._pluginContext };
      }
    }
    return {
      ...extraConfig,
      ...this.config
    };
  }
  updateConfig(config) {
    this.config = {
      ...this.getConfig(),
      ...config
    };
  }
  enableCheckMode() {
    this._checkMode = true;
  }
  get checkMode() {
    return this._checkMode;
  }
  useProfiler() {
    this.profiler = createProfiler();
    const now = new Date();
    const datetime = now.toISOString().split(".")[0];
    const datetimeNormalized = datetime.replace(/-|:/g, "");
    this.profilerOutput = `codegen-${datetimeNormalized}.json`;
  }
  getPluginContext() {
    return this._pluginContext;
  }
  async loadSchema(pointer) {
    const config = this.getConfig(defaultSchemaLoadOptions);
    if (this._graphqlConfig) {
      return addHashToSchema(
        this._graphqlConfig.getProject(this._project).loadSchema(pointer, "GraphQLSchema", config)
      );
    }
    return addHashToSchema(loadSchema(pointer, config));
  }
  async loadDocuments(pointer) {
    const config = this.getConfig(defaultDocumentsLoadOptions);
    if (this._graphqlConfig) {
      return addHashToDocumentFiles(this._graphqlConfig.getProject(this._project).loadDocuments(pointer, config));
    }
    return addHashToDocumentFiles(loadDocuments(pointer, config));
  }
}
function ensureContext(input) {
  return input instanceof CodegenContext ? input : new CodegenContext({ config: input });
}
function hashContent(content) {
  return createHash("sha256").update(content).digest("hex");
}
function hashSchema(schema) {
  return hashContent(print(getCachedDocumentNodeFromSchema(schema)));
}
function addHashToSchema(schemaPromise) {
  return schemaPromise.then((schema) => {
    if (!schema.extensions) {
      schema.extensions = {};
    }
    schema.extensions["hash"] = hashSchema(schema);
    return schema;
  });
}
function hashDocument(doc) {
  if (doc.rawSDL) {
    return hashContent(doc.rawSDL);
  }
  if (doc.document) {
    return hashContent(print(doc.document));
  }
  return null;
}
function addHashToDocumentFiles(documentFilesPromise) {
  return documentFilesPromise.then(
    (documentFiles) => documentFiles.map((doc) => {
      doc.hash = hashDocument(doc);
      return doc;
    })
  );
}
function shouldEmitLegacyCommonJSImports(config, outputPath) {
  const globalValue = config.emitLegacyCommonJSImports === void 0 ? true : !!config.emitLegacyCommonJSImports;
  return globalValue;
}

const isESMModule = typeof __dirname === "string" === false;
const makeDefaultLoader = (from) => {
  if (fs.statSync(from).isDirectory()) {
    from = path.join(from, "__fake.js");
  }
  const relativeRequire = createRequire(from);
  return async (mod) => {
    return isESMModule ? import(mod) : import(relativeRequire.resolve(mod));
  };
};
function createCache() {
  const cache = /* @__PURE__ */ new Map();
  return function ensure(namespace, key, factory) {
    const cacheKey = `${namespace}:${key}`;
    const cachedValue = cache.get(cacheKey);
    if (cachedValue) {
      return cachedValue;
    }
    const value = factory();
    cache.set(cacheKey, value);
    return value;
  };
}
async function executeCodegen(input) {
  const context = ensureContext(input);
  const config = context.getConfig();
  const pluginContext = context.getPluginContext();
  const result = [];
  let rootConfig = {};
  let rootSchemas;
  let rootDocuments;
  const generates = {};
  const cache = createCache();
  function wrapTask(task, source, taskName, ctx) {
    return () => {
      return context.profiler.run(async () => {
        try {
          await Promise.resolve().then(() => task());
        } catch (error) {
          if (source && !(error instanceof GraphQLError)) {
            error.source = source;
          }
          ctx.errors.push(error);
          throw error;
        }
      }, taskName);
    };
  }
  async function normalize() {
    const requireExtensions = normalizeInstanceOrArray(config.require);
    const loader = makeDefaultLoader(context.cwd);
    for (const mod of requireExtensions) {
      await loader(mod);
    }
    rootConfig = config.config || {};
    rootSchemas = normalizeInstanceOrArray(config.schema);
    rootDocuments = normalizeInstanceOrArray(config.documents);
    const generateKeys = Object.keys(config.generates || {});
    if (generateKeys.length === 0) {
      throw new Error(
        `Invalid Codegen Configuration! 

        Please make sure that your codegen config file contains the "generates" field, with a specification for the plugins you need.

        It should looks like that:

        schema:
          - my-schema.graphql
        generates:
          my-file.ts:
            - plugin1
            - plugin2
            - plugin3`
      );
    }
    for (const filename of generateKeys) {
      const output = generates[filename] = normalizeOutputParam(config.generates[filename]);
      if (!output.preset && (!output.plugins || output.plugins.length === 0)) {
        throw new Error(
          `Invalid Codegen Configuration! 

          Please make sure that your codegen config file has defined plugins list for output "${filename}".

          It should looks like that:

          schema:
            - my-schema.graphql
          generates:
            my-file.ts:
              - plugin1
              - plugin2
              - plugin3
          `
        );
      }
    }
    if (rootSchemas.length === 0 && Object.keys(generates).some(
      (filename) => !generates[filename].schema || Array.isArray(generates[filename].schema === "object") && generates[filename].schema.length === 0
    )) {
      throw new Error(
        `Invalid Codegen Configuration! 

        Please make sure that your codegen config file contains either the "schema" field
        or every generated file has its own "schema" field.

        It should looks like that:
        schema:
          - my-schema.graphql

        or:
        generates:
          path/to/output:
            schema: my-schema.graphql
      `
      );
    }
  }
  const isTest = process.env.NODE_ENV === "test";
  const tasks = new Listr(
    [
      {
        title: "Parse Configuration",
        task: () => normalize()
      },
      {
        title: "Generate outputs",
        task: (ctx, task) => {
          const generateTasks = Object.keys(generates).map((filename) => {
            const outputConfig = generates[filename];
            const hasPreset = !!outputConfig.preset;
            const title = `Generate to ${filename}`;
            return {
              title,
              task: async (_, subTask) => {
                let outputSchemaAst;
                let outputSchema;
                const outputFileTemplateConfig = outputConfig.config || {};
                let outputDocuments = [];
                const outputSpecificSchemas = normalizeInstanceOrArray(outputConfig.schema);
                let outputSpecificDocuments = normalizeInstanceOrArray(outputConfig.documents);
                const preset = hasPreset ? typeof outputConfig.preset === "string" ? await getPresetByName(outputConfig.preset, makeDefaultLoader(context.cwd)) : outputConfig.preset : null;
                if (preset) {
                  if (preset.prepareDocuments) {
                    outputSpecificDocuments = await preset.prepareDocuments(filename, outputSpecificDocuments);
                  }
                }
                return subTask.newListr(
                  [
                    {
                      title: "Load GraphQL schemas",
                      task: wrapTask(
                        async () => {
                          debugLog(`[CLI] Loading Schemas`);
                          const schemaPointerMap = {};
                          const allSchemaDenormalizedPointers = [...rootSchemas, ...outputSpecificSchemas];
                          for (const denormalizedPtr of allSchemaDenormalizedPointers) {
                            if (typeof denormalizedPtr === "string") {
                              schemaPointerMap[denormalizedPtr] = {};
                            } else if (typeof denormalizedPtr === "object") {
                              Object.assign(schemaPointerMap, denormalizedPtr);
                            }
                          }
                          const hash = JSON.stringify(schemaPointerMap);
                          const result2 = await cache("schema", hash, async () => {
                            const outputSchemaAst2 = await context.loadSchema(schemaPointerMap);
                            const outputSchema2 = getCachedDocumentNodeFromSchema(outputSchemaAst2);
                            return {
                              outputSchemaAst: outputSchemaAst2,
                              outputSchema: outputSchema2
                            };
                          });
                          outputSchemaAst = result2.outputSchemaAst;
                          outputSchema = result2.outputSchema;
                        },
                        filename,
                        `Load GraphQL schemas: ${filename}`,
                        ctx
                      )
                    },
                    {
                      title: "Load GraphQL documents",
                      task: wrapTask(
                        async () => {
                          debugLog(`[CLI] Loading Documents`);
                          const documentPointerMap = {};
                          const allDocumentsDenormalizedPointers = [...rootDocuments, ...outputSpecificDocuments];
                          for (const denormalizedPtr of allDocumentsDenormalizedPointers) {
                            if (typeof denormalizedPtr === "string") {
                              documentPointerMap[denormalizedPtr] = {};
                            } else if (typeof denormalizedPtr === "object") {
                              Object.assign(documentPointerMap, denormalizedPtr);
                            }
                          }
                          const hash = JSON.stringify(documentPointerMap);
                          const result2 = await cache("documents", hash, async () => {
                            const documents = await context.loadDocuments(documentPointerMap);
                            return {
                              documents
                            };
                          });
                          outputDocuments = result2.documents;
                        },
                        filename,
                        `Load GraphQL documents: ${filename}`,
                        ctx
                      )
                    },
                    {
                      title: "Generate",
                      task: wrapTask(
                        async () => {
                          debugLog(`[CLI] Generating output`);
                          const normalizedPluginsArray = normalizeConfig(outputConfig.plugins);
                          const pluginLoader = config.pluginLoader || makeDefaultLoader(context.cwd);
                          const pluginPackages = await Promise.all(
                            normalizedPluginsArray.map((plugin) => getPluginByName(Object.keys(plugin)[0], pluginLoader))
                          );
                          const pluginMap = Object.fromEntries(
                            pluginPackages.map((pkg, i) => {
                              const plugin = normalizedPluginsArray[i];
                              const name = Object.keys(plugin)[0];
                              return [name, pkg];
                            })
                          );
                          const mergedConfig = {
                            ...rootConfig,
                            ...typeof outputFileTemplateConfig === "string" ? { value: outputFileTemplateConfig } : outputFileTemplateConfig,
                            emitLegacyCommonJSImports: shouldEmitLegacyCommonJSImports(config, filename)
                          };
                          const outputs = preset ? await context.profiler.run(
                            async () => preset.buildGeneratesSection({
                              baseOutputDir: filename,
                              presetConfig: outputConfig.presetConfig || {},
                              plugins: normalizedPluginsArray,
                              schema: outputSchema,
                              schemaAst: outputSchemaAst,
                              documents: outputDocuments,
                              config: mergedConfig,
                              pluginMap,
                              pluginContext,
                              profiler: context.profiler
                            }),
                            `Build Generates Section: ${filename}`
                          ) : [
                            {
                              filename,
                              plugins: normalizedPluginsArray,
                              schema: outputSchema,
                              schemaAst: outputSchemaAst,
                              documents: outputDocuments,
                              config: mergedConfig,
                              pluginMap,
                              pluginContext,
                              profiler: context.profiler
                            }
                          ];
                          const process2 = async (outputArgs) => {
                            const output = await codegen({
                              ...{
                                ...outputArgs,
                                emitLegacyCommonJSImports: shouldEmitLegacyCommonJSImports(config, outputArgs.filename)
                              },
                              cache
                            });
                            result.push({
                              filename: outputArgs.filename,
                              content: output,
                              hooks: outputConfig.hooks || {}
                            });
                          };
                          await context.profiler.run(() => Promise.all(outputs.map(process2)), `Codegen: ${filename}`);
                        },
                        filename,
                        `Generate: ${filename}`,
                        ctx
                      )
                    }
                  ],
                  {
                    exitOnError: true
                  }
                );
              },
              exitOnError: false,
              concurrent: cpus().length
            };
          });
          return task.newListr(generateTasks);
        }
      }
    ],
    {
      rendererOptions: {
        clearOutput: false,
        collapse: true
      },
      renderer: config.verbose ? "verbose" : "default",
      ctx: { errors: [] },
      rendererSilent: isTest || config.silent,
      exitOnError: true
    }
  );
  const executedContext = await tasks.run();
  if (config.debug) {
    printLogs();
  }
  if (executedContext.errors.length > 0) {
    const errors = executedContext.errors.map((subErr) => subErr.message || subErr.toString());
    const newErr = new AggregateError(executedContext.errors, `${errors.join("\n\n")}`);
    newErr.stack = `${newErr.stack}

${executedContext.errors.map((subErr) => subErr.stack).join("\n\n")}`;
    throw newErr;
  }
  return result;
}

const DEFAULT_HOOKS = {
  afterStart: [],
  beforeDone: [],
  onWatchTriggered: [],
  onError: [],
  afterOneFileWrite: [],
  afterAllFileWrite: [],
  beforeOneFileWrite: [],
  beforeAllFileWrite: []
};
function normalizeHooks(_hooks) {
  const keys = Object.keys({
    ...DEFAULT_HOOKS,
    ..._hooks
  });
  return keys.reduce((prev, hookName) => {
    if (typeof _hooks[hookName] === "string") {
      return {
        ...prev,
        [hookName]: [_hooks[hookName]]
      };
    }
    if (typeof _hooks[hookName] === "function") {
      return {
        ...prev,
        [hookName]: [_hooks[hookName]]
      };
    }
    if (Array.isArray(_hooks[hookName])) {
      return {
        ...prev,
        [hookName]: _hooks[hookName]
      };
    }
    return prev;
  }, {});
}
function execShellCommand(cmd) {
  return new Promise((resolve, reject) => {
    exec(
      cmd,
      {
        env: {
          ...process.env,
          PATH: `${process.env.PATH}${delimiter}${process.cwd()}${sep}node_modules${sep}.bin`
        }
      },
      (error, stdout, stderr) => {
        if (error) {
          reject(error);
          console.error(error);
        } else {
          debugLog(stdout || stderr);
          resolve(stdout || stderr);
        }
      }
    );
  });
}
async function executeHooks(hookName, scripts = [], args = []) {
  debugLog(`Running lifecycle hook "${hookName}" scripts...`);
  const quotedArgs = quote(args);
  for (const script of scripts) {
    if (typeof script === "string") {
      debugLog(`Running lifecycle hook "${hookName}" script: ${script} with args: ${quotedArgs}...`);
      await execShellCommand(`${script} ${quotedArgs}`);
    } else {
      debugLog(`Running lifecycle hook "${hookName}" script: ${script.name} with args: ${args.join(" ")}...`);
      await script(...args);
    }
  }
}
const lifecycleHooks = (_hooks = {}) => {
  const hooks = normalizeHooks(_hooks);
  return {
    afterStart: async () => executeHooks("afterStart", hooks.afterStart),
    onWatchTriggered: async (event, path) => executeHooks("onWatchTriggered", hooks.onWatchTriggered, [event, path]),
    onError: async (error) => executeHooks("onError", hooks.onError, [error]),
    afterOneFileWrite: async (path) => executeHooks("afterOneFileWrite", hooks.afterOneFileWrite, [path]),
    afterAllFileWrite: async (paths) => executeHooks("afterAllFileWrite", hooks.afterAllFileWrite, paths),
    beforeOneFileWrite: async (path) => executeHooks("beforeOneFileWrite", hooks.beforeOneFileWrite, [path]),
    beforeAllFileWrite: async (paths) => executeHooks("beforeAllFileWrite", hooks.beforeAllFileWrite, paths),
    beforeDone: async () => executeHooks("beforeDone", hooks.beforeDone)
  };
};

function log$1(msg) {
  getLogger().info(`  ${msg}`);
}
function emitWatching() {
  log$1(`${logSymbols.info} Watching for changes...`);
}
const createWatcher = (initalContext, onNext) => {
  debugLog(`[Watcher] Starting watcher...`);
  let config = initalContext.getConfig();
  const files = [initalContext.filepath].filter((a) => a);
  const documents = normalizeInstanceOrArray(config.documents);
  const schemas = normalizeInstanceOrArray(config.schema);
  Object.keys(config.generates).map((filename) => normalizeOutputParam(config.generates[filename])).forEach((conf) => {
    schemas.push(...normalizeInstanceOrArray(conf.schema));
    documents.push(...normalizeInstanceOrArray(conf.documents));
  });
  if (documents) {
    documents.forEach((doc) => {
      if (typeof doc === "string") {
        files.push(doc);
      } else {
        files.push(...Object.keys(doc));
      }
    });
  }
  schemas.forEach((schema) => {
    if (isGlob(schema) || isValidPath(schema)) {
      files.push(schema);
    }
  });
  if (typeof config.watch !== "boolean") {
    files.push(...normalizeInstanceOrArray(config.watch));
  }
  let watcher;
  const runWatcher = async () => {
    const chokidar = await import('chokidar');
    let isShutdown = false;
    const debouncedExec = debounce(() => {
      if (!isShutdown) {
        executeCodegen(initalContext).then(onNext, () => Promise.resolve()).then(() => emitWatching());
      }
    }, 100);
    emitWatching();
    const ignored = [];
    Object.keys(config.generates).map((filename) => ({ filename, config: normalizeOutputParam(config.generates[filename]) })).forEach((entry) => {
      if (entry.config.preset) {
        const extension = entry.config.presetConfig && entry.config.presetConfig.extension;
        if (extension) {
          ignored.push(join(entry.filename, "**", "*" + extension));
        }
      } else {
        ignored.push(entry.filename);
      }
    });
    watcher = chokidar.watch(files, {
      persistent: true,
      ignoreInitial: true,
      followSymlinks: true,
      cwd: process.cwd(),
      disableGlobbing: false,
      usePolling: config.watchConfig?.usePolling,
      interval: config.watchConfig?.interval,
      depth: 99,
      awaitWriteFinish: true,
      ignorePermissionErrors: false,
      atomic: true,
      ignored
    });
    debugLog(`[Watcher] Started`);
    const shutdown = () => {
      isShutdown = true;
      debugLog(`[Watcher] Shutting down`);
      log$1(`Shutting down watch...`);
      watcher.close();
      lifecycleHooks(config.hooks).beforeDone();
    };
    watcher.on("all", async (eventName, path) => {
      lifecycleHooks(config.hooks).onWatchTriggered(eventName, path);
      debugLog(`[Watcher] triggered due to a file ${eventName} event: ${path}`);
      const fullPath = join(process.cwd(), path);
      try {
        delete require.cache[fullPath];
      } catch (err) {
      }
      if (eventName === "change" && config.configFilePath && fullPath === config.configFilePath) {
        log$1(`${logSymbols.info} Config file has changed, reloading...`);
        const context = await loadContext(config.configFilePath);
        const newParsedConfig = context.getConfig();
        newParsedConfig.watch = config.watch;
        newParsedConfig.silent = config.silent;
        newParsedConfig.overwrite = config.overwrite;
        newParsedConfig.configFilePath = config.configFilePath;
        config = newParsedConfig;
        initalContext.updateConfig(config);
      }
      debouncedExec();
    });
    process.once("SIGINT", shutdown);
    process.once("SIGTERM", shutdown);
  };
  return new Promise((resolve, reject) => {
    executeCodegen(initalContext).then(onNext, () => Promise.resolve()).then(runWatcher).catch((err) => {
      watcher.close();
      reject(err);
    });
  });
};

const { writeFile: fsWriteFile, readFile: fsReadFile, stat: fsStat } = promises;
function writeFile(filepath, content) {
  return fsWriteFile(filepath, content);
}
function readFile(filepath) {
  return fsReadFile(filepath, "utf-8");
}
async function fileExists(filePath) {
  try {
    return (await fsStat(filePath)).isFile();
  } catch (err) {
    return false;
  }
}
function unlinkFile(filePath, cb) {
  unlink(filePath, cb);
}

const hash = (content) => createHash("sha1").update(content).digest("base64");
async function generate(input, saveToFile = true) {
  const context = ensureContext(input);
  const config = context.getConfig();
  await context.profiler.run(() => lifecycleHooks(config.hooks).afterStart(), "Lifecycle: afterStart");
  let previouslyGeneratedFilenames = [];
  function removeStaleFiles(config2, generationResult) {
    const filenames = generationResult.map((o) => o.filename);
    const staleFilenames = previouslyGeneratedFilenames.filter((f) => !filenames.includes(f));
    staleFilenames.forEach((filename) => {
      if (shouldOverwrite(config2, filename)) {
        return unlinkFile(filename, (err) => {
          const prettyFilename = filename.replace(`${input.cwd || process.cwd()}/`, "");
          if (err) {
            debugLog(`Cannot remove stale file: ${prettyFilename}
${err}`);
          } else {
            debugLog(`Removed stale file: ${prettyFilename}`);
          }
        });
      }
    });
    previouslyGeneratedFilenames = filenames;
  }
  const recentOutputHash = /* @__PURE__ */ new Map();
  async function writeOutput(generationResult) {
    if (!saveToFile) {
      return generationResult;
    }
    if (config.watch) {
      removeStaleFiles(config, generationResult);
    }
    await context.profiler.run(async () => {
      await lifecycleHooks(config.hooks).beforeAllFileWrite(generationResult.map((r) => r.filename));
    }, "Lifecycle: beforeAllFileWrite");
    await context.profiler.run(
      () => Promise.all(
        generationResult.map(async (result) => {
          const exists = await fileExists(result.filename);
          if (!shouldOverwrite(config, result.filename) && exists) {
            return;
          }
          const content = result.content || "";
          const currentHash = hash(content);
          let previousHash = recentOutputHash.get(result.filename);
          if (!previousHash && exists) {
            previousHash = hash(await readFile(result.filename));
          }
          if (previousHash && currentHash === previousHash) {
            debugLog(`Skipping file (${result.filename}) writing due to indentical hash...`);
            return;
          } else if (context.checkMode) {
            context.checkModeStaleFiles.push(result.filename);
            return;
          }
          if (content.length === 0) {
            return;
          }
          recentOutputHash.set(result.filename, currentHash);
          const basedir = dirname(result.filename);
          await lifecycleHooks(result.hooks).beforeOneFileWrite(result.filename);
          await lifecycleHooks(config.hooks).beforeOneFileWrite(result.filename);
          await mkdirp(basedir);
          const absolutePath = isAbsolute(result.filename) ? result.filename : join(input.cwd || process.cwd(), result.filename);
          await writeFile(absolutePath, result.content);
          await lifecycleHooks(result.hooks).afterOneFileWrite(result.filename);
          await lifecycleHooks(config.hooks).afterOneFileWrite(result.filename);
        })
      ),
      "Write files"
    );
    await context.profiler.run(
      () => lifecycleHooks(config.hooks).afterAllFileWrite(generationResult.map((r) => r.filename)),
      "Lifecycle: afterAllFileWrite"
    );
    return generationResult;
  }
  if (config.watch) {
    return createWatcher(context, writeOutput);
  }
  const outputFiles = await context.profiler.run(() => executeCodegen(context), "executeCodegen");
  await context.profiler.run(() => writeOutput(outputFiles), "writeOutput");
  await context.profiler.run(() => lifecycleHooks(config.hooks).beforeDone(), "Lifecycle: beforeDone");
  if (context.profilerOutput) {
    await writeFile(join(context.cwd, context.profilerOutput), JSON.stringify(context.profiler.collect()));
  }
  return outputFiles;
}
function shouldOverwrite(config, outputPath) {
  const globalValue = config.overwrite === void 0 ? true : !!config.overwrite;
  const outputConfig = config.generates[outputPath];
  if (!outputConfig) {
    debugLog(`Couldn't find a config of ${outputPath}`);
    return globalValue;
  }
  if (isConfiguredOutput(outputConfig) && typeof outputConfig.overwrite === "boolean") {
    return outputConfig.overwrite;
  }
  return globalValue;
}
function isConfiguredOutput(output) {
  return typeof output.plugins !== "undefined";
}

var Tags = /* @__PURE__ */ ((Tags2) => {
  Tags2["client"] = "Browser";
  Tags2["node"] = "Node";
  Tags2["typescript"] = "TypeScript";
  Tags2["flow"] = "Flow";
  Tags2["angular"] = "Angular";
  Tags2["stencil"] = "Stencil";
  Tags2["react"] = "React";
  Tags2["vue"] = "Vue";
  Tags2["graphqlRequest"] = "graphqlRequest";
  return Tags2;
})(Tags || {});

async function getLatestVersion(packageName) {
  return fetch(`https://unpkg.com/${packageName}/package.json`).then((res) => res.json()).then((pkg) => pkg.version);
}

function jsObjectToBabelObjectExpression(obj) {
  const objExp = t.objectExpression([]);
  Object.entries(obj).forEach(([key, val]) => {
    if (Array.isArray(val)) {
      objExp.properties.push(
        t.objectProperty(
          /^[a-zA-Z0-9]+$/.test(key) ? t.identifier(key) : t.stringLiteral(key),
          t.arrayExpression(
            val.map((v) => typeof v === "object" ? jsObjectToBabelObjectExpression(v) : t.valueToNode(v))
          )
        )
      );
    } else {
      objExp.properties.push(
        t.objectProperty(
          /^[a-zA-Z0-9]+$/.test(key) ? t.identifier(key) : t.stringLiteral(key),
          typeof val === "object" ? jsObjectToBabelObjectExpression(val) : t.valueToNode(val)
        )
      );
    }
  });
  return objExp;
}
async function writeConfig(answers, config) {
  const YAML = await import('json-to-pretty-yaml').then((m) => "default" in m ? m.default : m);
  const ext = answers.config.toLocaleLowerCase().split(".")[1];
  const fullPath = resolve(process.cwd(), answers.config);
  const relativePath = relative(process.cwd(), answers.config);
  let content;
  if (ext === "ts") {
    const buildRequire = template.statement(`%%config%%`);
    const ast = buildRequire({
      config: jsObjectToBabelObjectExpression(config)
    });
    content = `
import type { CodegenConfig } from '@graphql-codegen/cli';

const config: CodegenConfig = ${generate$1(ast).code.replace(/\(|\)/g, "")}

export default config;
`;
  } else {
    content = ext === "json" ? JSON.stringify(config) : YAML.stringify(config);
  }
  writeFileSync(fullPath, content, {
    encoding: "utf-8"
  });
  return {
    relativePath,
    fullPath
  };
}
async function writePackage(answers, configLocation) {
  const pkgPath = resolve(process.cwd(), "package.json");
  const pkgContent = readFileSync(pkgPath, {
    encoding: "utf-8"
  });
  const pkg = JSON.parse(pkgContent);
  const { indent } = detectIndent(pkgContent);
  if (!pkg.scripts) {
    pkg.scripts = {};
  }
  pkg.scripts[answers.script] = `graphql-codegen --config ${configLocation}`;
  if (!pkg.devDependencies) {
    pkg.devDependencies = {};
  }
  await Promise.all(
    (answers.plugins || []).map(async (plugin) => {
      pkg.devDependencies[plugin.package] = await getLatestVersion(plugin.package);
    })
  );
  if (answers.introspection) {
    pkg.devDependencies["@graphql-codegen/introspection"] = await getLatestVersion("@graphql-codegen/introspection");
  }
  pkg.devDependencies["@graphql-codegen/cli"] = await getLatestVersion("@graphql-codegen/cli");
  if (answers.targets.includes(Tags.client)) {
    pkg.devDependencies["@graphql-codegen/client-preset"] = await getLatestVersion("@graphql-codegen/client-preset");
  }
  writeFileSync(pkgPath, JSON.stringify(pkg, null, indent));
}
function bold(str) {
  return chalk.bold(str);
}
function grey(str) {
  return chalk.grey(str);
}
function italic(str) {
  return chalk.italic(str);
}

const plugins = [
  {
    name: `TypeScript ${italic("(required by other typescript plugins)")}`,
    package: "@graphql-codegen/typescript",
    value: "typescript",
    pathInRepo: "typescript/typescript",
    available: hasTag(Tags.typescript),
    shouldBeSelected: (tags) => oneOf(tags, Tags.angular, Tags.stencil) || allOf(tags, Tags.typescript, Tags.react) || noneOf(tags, Tags.flow),
    defaultExtension: ".ts"
  },
  {
    name: `TypeScript Operations ${italic("(operations and fragments)")}`,
    package: "@graphql-codegen/typescript-operations",
    value: "typescript-operations",
    pathInRepo: "typescript/operations",
    available: (tags) => allOf(tags, Tags.client, Tags.typescript) || hasTag(Tags.stencil)(tags),
    shouldBeSelected: (tags) => oneOf(tags, Tags.angular, Tags.stencil) || allOf(tags, Tags.typescript, Tags.react),
    defaultExtension: ".ts"
  },
  {
    name: `TypeScript Resolvers ${italic("(strongly typed resolve functions)")}`,
    package: "@graphql-codegen/typescript-resolvers",
    value: "typescript-resolvers",
    pathInRepo: "typescript/resolvers",
    available: (tags) => allOf(tags, Tags.node, Tags.typescript),
    shouldBeSelected: (tags) => noneOf(tags, Tags.flow),
    defaultExtension: ".ts"
  },
  {
    name: `Flow ${italic("(required by other flow plugins)")}`,
    package: "@graphql-codegen/flow",
    value: "flow",
    pathInRepo: "flow/flow",
    available: hasTag(Tags.flow),
    shouldBeSelected: (tags) => noneOf(tags, Tags.typescript),
    defaultExtension: ".js"
  },
  {
    name: `Flow Operations ${italic("(operations and fragments)")}`,
    package: "@graphql-codegen/flow-operations",
    value: "flow-operations",
    pathInRepo: "flow/operations",
    available: (tags) => allOf(tags, Tags.client, Tags.flow),
    shouldBeSelected: (tags) => noneOf(tags, Tags.typescript),
    defaultExtension: ".js"
  },
  {
    name: `Flow Resolvers ${italic("(strongly typed resolve functions)")}`,
    package: "@graphql-codegen/flow-resolvers",
    value: "flow-resolvers",
    pathInRepo: "flow/resolvers",
    available: (tags) => allOf(tags, Tags.node, Tags.flow),
    shouldBeSelected: (tags) => noneOf(tags, Tags.typescript),
    defaultExtension: ".js"
  },
  {
    name: `TypeScript Stencil Apollo ${italic("(typed components)")}`,
    package: "@graphql-codegen/typescript-stencil-apollo",
    value: "typescript-stencil-apollo",
    pathInRepo: "typescript/stencil-apollo",
    available: hasTag(Tags.stencil),
    shouldBeSelected: () => true,
    defaultExtension: ".tsx"
  },
  {
    name: `TypeScript MongoDB ${italic("(typed MongoDB objects)")}`,
    package: "@graphql-codegen/typescript-mongodb",
    value: "typescript-mongodb",
    pathInRepo: "typescript/mongodb",
    available: (tags) => allOf(tags, Tags.node, Tags.typescript),
    shouldBeSelected: () => false,
    defaultExtension: ".ts"
  },
  {
    name: `TypeScript GraphQL files modules ${italic("(declarations for .graphql files)")}`,
    package: "@graphql-codegen/typescript-graphql-files-modules",
    value: "typescript-graphql-files-modules",
    pathInRepo: "typescript/graphql-files-modules",
    available: (tags) => allOf(tags, Tags.client, Tags.typescript) || hasTag(Tags.stencil)(tags),
    shouldBeSelected: () => false,
    defaultExtension: ".ts"
  },
  {
    name: `TypeScript GraphQL document nodes ${italic("(embedded GraphQL document)")}`,
    package: "@graphql-codegen/typescript-document-nodes",
    value: "typescript-document-nodes",
    pathInRepo: "typescript/document-nodes",
    available: (tags) => allOf(tags, Tags.typescript) || hasTag(Tags.stencil)(tags),
    shouldBeSelected: () => false,
    defaultExtension: ".ts"
  },
  {
    name: `Introspection Fragment Matcher ${italic("(for Apollo Client)")}`,
    package: "@graphql-codegen/fragment-matcher",
    value: "fragment-matcher",
    pathInRepo: "other/fragment-matcher",
    available: (tags) => hasTag(Tags.client)(tags) || hasTag(Tags.angular)(tags) || hasTag(Tags.stencil)(tags),
    shouldBeSelected: () => false,
    defaultExtension: ".ts"
  },
  {
    name: `Urql Introspection ${italic("(for Urql Client)")}`,
    package: "@graphql-codegen/urql-introspection",
    value: "urql-introspection",
    pathInRepo: "other/urql-introspection",
    available: (tags) => hasTag(Tags.client)(tags) || hasTag(Tags.stencil)(tags),
    shouldBeSelected: () => false,
    defaultExtension: ".ts"
  },
  {
    name: `TypeScript Apollo Angular ${italic("(typed GQL services)")}`,
    package: "@graphql-codegen/typescript-apollo-angular",
    value: "typescript-apollo-angular",
    pathInRepo: "typescript/apollo-angular",
    available: hasTag(Tags.angular),
    shouldBeSelected: () => true,
    defaultExtension: ".ts"
  }
];
function hasTag(tag) {
  return (tags) => tags.includes(tag);
}
function oneOf(list, ...items) {
  return list.some((i) => items.includes(i));
}
function noneOf(list, ...items) {
  return !list.some((i) => items.includes(i));
}
function allOf(list, ...items) {
  return items.every((i) => list.includes(i));
}

function getQuestions(possibleTargets) {
  return [
    {
      type: "list",
      name: "targets",
      message: `What type of application are you building?`,
      choices: getApplicationTypeChoices(possibleTargets),
      validate: (targets) => targets.length > 0,
      default: getApplicationTypeChoices(possibleTargets).findIndex((c) => c.checked)
    },
    {
      type: "input",
      name: "schema",
      message: "Where is your schema?:",
      suffix: grey(" (path or url)"),
      default: "http://localhost:4000",
      validate: (str) => str.length > 0
    },
    {
      type: "input",
      name: "documents",
      message: "Where are your operations and fragments?:",
      when: (answers) => {
        answers.targets = normalizeTargets(answers.targets);
        return answers.targets.includes(Tags.client) || answers.targets.includes(Tags.angular) || answers.targets.includes(Tags.stencil);
      },
      default: getDocumentsDefaultValue,
      validate: (str) => str.length > 0
    },
    {
      type: "checkbox",
      name: "plugins",
      when: (answers) => {
        answers.targets = normalizeTargets(answers.targets);
        return !answers.targets.includes(Tags.client);
      },
      message: "Pick plugins:",
      choices: getPluginChoices,
      validate: (plugins2) => plugins2.length > 0
    },
    {
      type: "input",
      name: "output",
      message: "Where to write the output:",
      default: getOutputDefaultValue,
      validate: (str) => str.length > 0
    },
    {
      type: "confirm",
      name: "introspection",
      default: false,
      message: "Do you want to generate an introspection file?"
    },
    {
      type: "input",
      name: "config",
      message: "How to name the config file?",
      default: (answers) => answers.targets.includes(Tags.client) || answers.targets.includes(Tags.typescript) || answers.targets.includes(Tags.angular) ? "codegen.ts" : "codegen.yml",
      validate: (str) => {
        const isNotEmpty = str.length > 0;
        const hasCorrectExtension = ["json", "yml", "yaml", "js", "ts"].some(
          (ext) => str.toLocaleLowerCase().endsWith(`.${ext}`)
        );
        return isNotEmpty && hasCorrectExtension;
      }
    },
    {
      type: "input",
      name: "script",
      default: "codegen",
      message: "What script in package.json should run the codegen?",
      validate: (str) => str.length > 0
    }
  ];
}
function getApplicationTypeChoices(possibleTargets) {
  function withFlowOrTypescript(tags) {
    if (possibleTargets.TypeScript) {
      tags.push(Tags.typescript);
    } else if (possibleTargets.Flow) {
      tags.push(Tags.flow);
    } else if (possibleTargets.Node) {
      tags.push(Tags.typescript);
      tags.push(Tags.flow);
    }
    return tags;
  }
  return [
    {
      name: "Backend - API or server",
      key: "backend",
      value: withFlowOrTypescript([Tags.node]),
      checked: possibleTargets.Node
    },
    {
      name: "Application built with Angular",
      key: "angular",
      value: [Tags.angular],
      checked: possibleTargets.Angular
    },
    {
      name: "Application built with React",
      key: "react",
      value: withFlowOrTypescript([Tags.react, Tags.client]),
      checked: possibleTargets.React
    },
    {
      name: "Application built with Stencil",
      key: "stencil",
      value: [Tags.stencil, Tags.typescript],
      checked: possibleTargets.Stencil
    },
    {
      name: "Application built with Vue",
      key: "vue",
      value: [Tags.vue, Tags.client],
      checked: possibleTargets.Vue
    },
    {
      name: "Application using graphql-request",
      key: "graphqlRequest",
      value: [Tags.graphqlRequest, Tags.client],
      checked: possibleTargets.graphqlRequest
    },
    {
      name: "Application built with other framework or vanilla JS",
      key: "client",
      value: [Tags.typescript, Tags.flow],
      checked: possibleTargets.Browser && !possibleTargets.Angular && !possibleTargets.React && !possibleTargets.Stencil
    }
  ];
}
function getPluginChoices(answers) {
  return plugins.filter((p) => p.available(answers.targets)).map((p) => {
    return {
      name: p.name,
      value: p,
      checked: p.shouldBeSelected(answers.targets)
    };
  });
}
function normalizeTargets(targets) {
  return [].concat(...targets);
}
function getOutputDefaultValue(answers) {
  if (answers.targets.includes(Tags.client)) {
    return "src/gql";
  }
  if (answers.plugins.some((plugin) => plugin.defaultExtension === ".tsx")) {
    return "src/generated/graphql.tsx";
  }
  if (answers.plugins.some((plugin) => plugin.defaultExtension === ".ts")) {
    return "src/generated/graphql.ts";
  }
  return "src/generated/graphql.js";
}
function getDocumentsDefaultValue(answers) {
  if (answers.targets.includes(Tags.vue)) {
    return "src/**/*.vue";
  } else if (answers.targets.includes(Tags.angular)) {
    return "src/**/*.ts";
  } else if (answers.targets.includes(Tags.client)) {
    return "src/**/*.tsx";
  } else {
    return "src/**/*.graphql";
  }
}

async function guessTargets() {
  const pkg = JSON.parse(
    readFileSync(resolve(process.cwd(), "package.json"), {
      encoding: "utf-8"
    })
  );
  const dependencies = Object.keys({
    ...pkg.dependencies,
    ...pkg.devDependencies
  });
  return {
    [Tags.angular]: isAngular(dependencies),
    [Tags.react]: isReact(dependencies),
    [Tags.stencil]: isStencil(dependencies),
    [Tags.vue]: isVue(dependencies),
    [Tags.client]: false,
    [Tags.node]: false,
    [Tags.typescript]: isTypescript(dependencies),
    [Tags.flow]: isFlow(dependencies),
    [Tags.graphqlRequest]: isGraphqlRequest(dependencies)
  };
}
function isAngular(dependencies) {
  return dependencies.includes("@angular/core");
}
function isReact(dependencies) {
  return dependencies.includes("react");
}
function isStencil(dependencies) {
  return dependencies.includes("@stencil/core");
}
function isVue(dependencies) {
  return dependencies.includes("vue") || dependencies.includes("nuxt");
}
function isTypescript(dependencies) {
  return dependencies.includes("typescript");
}
function isFlow(dependencies) {
  return dependencies.includes("flow");
}
function isGraphqlRequest(dependencies) {
  return dependencies.includes("graphql-request");
}

function log(...msgs) {
  console.log(...msgs);
}
async function init() {
  log(`
    Welcome to ${bold("GraphQL Code Generator")}!
    Answer few questions and we will setup everything for you.
  `);
  const possibleTargets = await guessTargets();
  const answers = await inquirer.prompt(getQuestions(possibleTargets));
  const config = {
    overwrite: true,
    schema: answers.schema,
    ...answers.targets.includes(Tags.client) || answers.targets.includes(Tags.angular) || answers.targets.includes(Tags.stencil) ? { documents: answers.documents } : {},
    generates: {
      [answers.output]: {
        ...answers.targets.includes(Tags.client) ? { preset: "client" } : {},
        plugins: answers.plugins ? answers.plugins.map((p) => p.value) : []
      }
    }
  };
  if (answers.introspection) {
    addIntrospection(config);
  }
  const { relativePath } = await writeConfig(answers, config);
  log(`Fetching latest versions of selected plugins...`);
  await writePackage(answers, relativePath);
  log(`
    Config file generated at ${bold(relativePath)}

      ${bold("$ npm install")}

    To install the plugins.

      ${bold(`$ npm run ${answers.script}`)}

    To run GraphQL Code Generator.
  `);
}
function addIntrospection(config) {
  config.generates["./graphql.schema.json"] = {
    plugins: ["introspection"]
  };
}

async function runCli(cmd) {
  await ensureGraphQlPackage();
  if (cmd === "init") {
    await init();
    return 0;
  }
  const context = await createContext();
  try {
    await generate(context);
    if (context.checkMode && context.checkModeStaleFiles.length > 0) {
      console.log(
        `The following stale files were detected:
${context.checkModeStaleFiles.map((file) => `  - ${file}
`)}`
      );
      return 1;
    }
    return 0;
  } catch (error) {
    await lifecycleHooks(context.getConfig().hooks).onError(error.toString());
    return 1;
  }
}
async function ensureGraphQlPackage() {
  try {
    await import('graphql');
  } catch (e) {
    throw new Error(
      `Unable to load "graphql" package. Please make sure to install "graphql" as a dependency! 

       To install "graphql", run:
         yarn add graphql
       Or, with NPM:
         npm install --save graphql`
    );
  }
}

export { CodegenContext as C, generateSearchPlaces as a, loadContext as b, buildOptions as c, createContext as d, executeCodegen as e, ensureContext as f, generate as g, ensureGraphQlPackage as h, init as i, CodegenExtension as j, findAndLoadGraphQLConfig as k, loadCodegenConfig as l, parseArgv as p, runCli as r, shouldEmitLegacyCommonJSImports as s, updateContextWithCliFlags as u };
